/*  @author : Michel Kinsy
 
 *  Copyright (c) 2012 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

#include <cstdlib>
#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <vector>
#include <stack>
#include <map>
using namespace std;

// Global variables
char *filename;             // input file varaible
int error = 0; 

typedef struct{
	string name;
	string parent; 
	map<string, string> variables;
	vector<string> childrens; 
	vector<string> lines; 
	vector<string> vars_lines;
} ELEM ;

void scanner(void); 

int main (int argc, char * const argv[]) {
	cout <<"--------------- Transforming .hc file to .c/c++ ------------------\n";
	if (argc < 2) return 1; 
	if(argc >= 2) filename = argv[1]; 
	scanner(); 
	if(error) cout<<"Some errors found during processing!!!\n"; 
	cout <<"----------------------- Processing done! -------------------------\n";
	return 0;
}

void add_types(vector<string>& types){
	types.push_back("float"); types.push_back("double");
	types.push_back("bool"); types.push_back("char");
	types.push_back("short"); types.push_back("long");
	types.push_back("int");types.push_back("void"); 
}

int find_element(vector<ELEM>& list, string name){
	int ans = -1; 
	for(int i = 0; i< list.size(); i++)
		if(list[i].name == name) ans = i; 
	return ans; 
}

int find_declaration (vector<ELEM>& list, string var, string name){
	int ans = -1; int k = -1; 
	map<string, string>::iterator iter;
	
	for(int i = 0; i< list.size(); i++)
		if(list[i].name == name){
			k = i; break; 
		}
	
	if( k != -1) {
		map<string, string>::iterator iter;
		iter = list[k].variables.find(var);
		if(iter != list[k].variables.end()) ans = k; 
		else ans = find_declaration (list, var, list[k].parent); 
	}
	
	return ans; 
}

void rearrange (vector<ELEM>& list, vector<string>& types){
	string lne, tok; 
	vector<string> nLines, fLines, cLines; 
	int lev, var, func; 
	for(int i = 0; i< list.size(); i++){
		lev = 0; func = 0; 
		nLines.clear(); fLines.clear(); cLines.clear();
		
		for(int j = 0; j< list[i].lines.size(); j++){
			istringstream t_list(list[i].lines[j]);
			t_list >> tok; 
			if(tok == "{") {
				istringstream n_list(list[i].lines[(j-1)]);
				n_list >> tok; var = 0; 
				for(int u = 0; u < types.size(); u++)
					if(tok == types[u]) var = 1; 
				while (n_list >> tok) if (tok =="=") var = 0; 
				
				if((lev == 1) && (var == 1)){
					func = 1;
					fLines.clear(); 
					fLines.push_back(list[i].lines[(j-1)]); 
				}
				lev++;
				
				if(func == 1) {
					int u = 0; string t_str = list[i].lines[j-1]; 
					while (t_str[u] !='\0') {
						if ((t_str[u] == '(') || (t_str[u] == ')') || (t_str[u] == ',')) t_str[u]= ' '; 	
						u++;
					}
					istringstream m_list(t_str); 
					tok =""; m_list >> tok; m_list >> tok;
					if(tok == "main") {
						if(nLines.size() > 0) nLines[nLines.size()-1] = "\tmain();"; 
					}
					else if(nLines.size() > 0) nLines[nLines.size()-1] = ""; 
				}
			}
			
			if(func == 1) fLines.push_back(list[i].lines[j]);
			else nLines.push_back(list[i].lines[j]);
			
			if(tok == "}") {
				lev--;
				if(lev == 1) {
					func = 0;
					for(int u = 0; u < fLines.size(); u++)
						cLines.push_back(fLines[u]); 
				}
			}
		}
		
		for(int u = 0; u < nLines.size(); u++)
			cLines.push_back(nLines[u]); 
		
		list[i].lines.clear(); 
		for(int u = 0; u < cLines.size(); u++)
			list[i].lines.push_back(cLines[u]); 
	}
}

int letter (char c) {
	int ans = 1; 
	if ((c < 48) || ((c > 57) && (c < 65)) || ((c > 90) && (c < 95)) || (c > 122)) ans = 0; 
	return ans; 
}

/* Scanning process and error cheching */
void scanner(void)
{
	stringstream ss;
	string new_file;
	
	string line, token, core, thread; 
	vector<string> lines, cur_lines; 
	vector<int> cores; 
	int index, len;
	int count = 0, looking = 0, normal = 1; 
	int cur_lev =  -1, level = -1, current = 0; 
	
	stack<int> stk; 
	vector<ELEM> elem_list;
	vector<string> types; add_types(types);  
	ELEM root; 
	root.name = "root"; 
	root.parent = "*"; 
	
	ifstream from;
	from.open(filename, ios::in);   /* open filename for reading */
	if(from.is_open())
		cin.rdbuf(from.rdbuf());  /* Redirect cin to the file reader buffer */
	else
	{
		cout << "Unable to open file.\n";  /* Explain the I/O error */
		exit(1);
	}
	
	ss << filename; ss >> new_file; 
	index = 0; len = new_file.length();
	while (filename[index] !='.') index++;
	if ((new_file.substr((index+1),(len-1)) == "c") || 
		(new_file.substr((index+1),(len-1)) == "cpp") ||
		(new_file.substr((index+1),(len-1)) == "cc")   ) {
		cout<<"C/C++ file No transformation need.\n"; 
		exit(1); 
	}
	new_file = new_file.substr(0,(index+1)) + new_file.substr((index+2),(len-1));
	
	
	/* Extracting the lines from the input file */
	cout<<"Start step 1/3 ... \n";
	remove("_temp.her"); 
	
	ofstream temp0 ("_temp.her", ios::out);
	while(getline(cin, line)) {	
		int u = 0; string t_line = line; index = 0; 
		while ((index = t_line.find("(", index)) != string::npos) {
			t_line.replace(index, 1, " ("); index+=2;
		}
		istringstream for_list(t_line); token = ""; 
		while(for_list >> token) if(token == "for") u = -1; 
		
		index = 0;
		if (u == 0) {
			while ((index = line.find(";", index)) != string::npos) {
				temp0<<line.substr(u, (index-u+1)) <<endl;
				u = (index+1); index+=1;
			}
		}
		if ((u == 0) || (u == -1)) temp0<<line<<endl;
	}
	temp0.close(); from.close(); 
	
	
	from.open("_temp.her", ios::in);   /* open filename for reading */
	if(from.is_open())
		cin.rdbuf(from.rdbuf());  /* Redirect cin to the file reader buffer */
	else
	{
		cout << "Temporary created file not found.\n";  /* Explain the I/O error */
		exit(1);
	}
	
	remove("__temp.her");
	ofstream temp ("__temp.her", ios::out);
	while(getline(cin, line)) {		
		index = 0;
		while ((index = line.find("]", index)) != string::npos) {
			line.replace(index, 1, "] ");
			index+=1;
		}
		index = 0;
		while ((index = line.find("[", index)) != string::npos) {
			line.replace(index, 1, " [");
			index+=2;
		}
		
		index = 0; 
		if ((index = line.find("=", index)) != string::npos) {
			temp<<line<<endl;
		}
		else {
			index = 0; 
			while ((index = line.find("{", index)) != string::npos) {
				line.replace(index, 1, "\n{\n");
				index+=2;
			}
			index = 0; 
			while ((index = line.find("}", index)) != string::npos) {
				line.replace(index, 1, "\n}\n");
				index+=2;
			}
			temp<<line<<endl;
		}
	}
	temp.close(); from.close();
	
	
	cout<<"Step 1 done ... \n";
	
	ofstream pos (new_file.c_str(), ios::out);
	from.open("__temp.her", ios::in);   /* open filename for reading */
	if(from.is_open())
		cin.rdbuf(from.rdbuf());  /* Redirect cin to the file reader buffer */
	else
	{
		cout << "Temporary created file not found.\n";  /* Explain the I/O error */
		exit(1);
	}
	
	while(getline(cin, line)) {
		istringstream token_list(line); count = 0; 
		while (token_list >> token) count++; 
		if(count > 0) lines.push_back(line); 
	}
	from.close(); remove("_temp.her"); remove("__temp.her"); 
	
	for(int i = 0; i< lines.size(); i++) {
		line = lines[i]; 
		
		int u = 0; 
		while (line[u] !='\0') {
			if (line[u]== ';')line[u]= ' '; u++;
		}
		
		istringstream token_list(line);
		token = ""; 
		u = 0;
		while(token_list >> token){
			if (token =="typedef") looking=1;
			if ((token =="}") && (looking == 1)) looking = 3;
			else if (looking == 3) {
				looking = 4; 
			}
			u++; 
		}
		
		if((u == 3) && (looking == 1)) {
			types.push_back(token); 
			looking = 0; 
		}
		if((u == 1) && (looking == 4)) {
			types.push_back(token); 
			looking = 0; 
		}
	}
	
	if (lines.size() < 1) {
		cout << "No valid lines in file.\n";  
		exit(1);
	}
	
	istringstream token_list(lines[0]);
	token = ""; token_list >> token; 
	
	if (token != "#heracles") {
		cout << "Not a valid heracles program (First line must be: #heracles) \n"; 
		remove(new_file.c_str()); 
		exit(1);
	}
	
	for(int i = 1; i< lines.size(); i++)
		cur_lines.push_back(lines[i]);
	
	
	root.variables["main_lock"] = "core_0_main_lock"; 
	root.vars_lines.push_back("int core_0_main_lock; ");
	
	root.lines.push_back("// Barriers and locks implementations"); 
	root.lines.push_back("void start(void){"); 
	root.lines.push_back("\tmain_lock = 1;"); 
	root.lines.push_back("}"); 
	root.lines.push_back("\n"); 
	root.lines.push_back("void start_check (int wait_count) {"); 
	root.lines.push_back("\tint i, a = 0; "); 
	root.lines.push_back("\twhile (main_lock != 1) {"); 
	root.lines.push_back("\t\ta = 0; "); 
	root.lines.push_back("\t\tfor(i = 0; i < wait_count; i++) a++; "); 
	root.lines.push_back("\t}"); 
	root.lines.push_back("}"); 
	root.lines.push_back("\n");
	root.lines.push_back("void check_lock (int* lock, int owner, int wait_count) {"); 
	root.lines.push_back("\tint i, a = 0; "); 
	root.lines.push_back("\twhile ((*lock) != owner) {"); 
	root.lines.push_back("\t\ta = 0; "); 
	root.lines.push_back("\t\tfor(i = 0; i < wait_count; i++) a++; "); 
	root.lines.push_back("\t}"); 
	root.lines.push_back("}"); 
	root.lines.push_back("\n");
	root.lines.push_back("void locking(int* lock, int owner) {"); 
	root.lines.push_back("\tif ((*lock) == 0)"); 
	root.lines.push_back("\t\t(*lock) = owner;"); 
	root.lines.push_back("}"); 
	root.lines.push_back("\n");
	root.lines.push_back("void release(int* lock, int owner) {"); 
	root.lines.push_back("\tif ((*lock) == owner) (*lock) = 0;"); 
	root.lines.push_back("}"); 
	root.lines.push_back("\n"); 
	root.lines.push_back("void clear(int* lock) {"); 
	root.lines.push_back("\t(*lock) = 0;"); 
	root.lines.push_back("}"); 
	root.lines.push_back("\n");
	root.lines.push_back("void set_barrier(int* sema, int num_elem) {");
	root.lines.push_back("\t(*sema) = num_elem;");
	root.lines.push_back("}");
	root.lines.push_back("\n");
	root.lines.push_back("void check_barrier(int* sema, int wait_count) {");
	root.lines.push_back("\tint i, a = 0; ");
	root.lines.push_back("\twhile ((*sema) != 0) {");
	root.lines.push_back("\ta = 0; ");
	root.lines.push_back("\tfor(i = 0; i < wait_count; i++) a++;");
	root.lines.push_back("\t}");
	root.lines.push_back("}");
	root.lines.push_back("\n");
	root.lines.push_back("void clear_barrier(int* sema) {");
	root.lines.push_back("\tif ((*sema) > 0)(*sema)--;");
	root.lines.push_back("}");
	root.lines.push_back("\n"); 
	root.lines.push_back("int h_print(int arg){"); 
	root.lines.push_back("\treturn arg;"); 
	root.lines.push_back("}"); 
	root.lines.push_back("\n"); 
	
	elem_list.push_back(root);
	
	cout<<"Start step 2/3 ... \n";
	
	for(int i = 0; i< cur_lines.size(); i++){
		line = cur_lines[i];  
		istringstream token_list(line);
		token = ""; token_list >> token;
		
		if (token =="#pragma") {
			token_list >> token; 
			if((token !="Heracles") && (token !="heracles")) error = 1;
			token_list >> token; 
			int type = 0; 
			if ((token =="core") || (token =="Core")) type = 1; 
			else if ((token =="thread") || (token =="Thread")) type = 2;
			else error = 1;
			token_list >> token; 
			if (type == 1) core = token; 
			if (type == 2) {
				thread = "T"+token;
				line = "void  core_" + core +"_work_"+ thread+" (void)";
				cur_lines[i] = line;
			}
 
		}
	}
	
	level = 0; 
	for(int i = 0; i< cur_lines.size(); i++){
		line = cur_lines[i]; normal = 1; 
		istringstream token_list(line);
		token = ""; token_list >> token;
		
		if (token =="#pragma") {
			token_list >> token; 
			if((token !="Heracles") && (token !="heracles")) error = 1;
			token_list >> token; 
			int type = 0; 
			if ((token == "core") || (token =="Core")) type = 1; 
			else if ((token !="thread") && (token!="Thread")) error = 1;
			token_list >> token; if (type == 1) core = token; 
			
			if (type == 1) {
				ELEM new_elem; new_elem.name = core; 
				new_elem.parent = elem_list[current].name;
				string str = "void  core_" + core +"_work (void)"; 
				new_elem.lines.push_back(str); 
				elem_list[current].childrens.push_back(core); 
				
				if (find_element(elem_list, core) == -1) 
					elem_list.push_back(new_elem);
				
				current = find_element(elem_list, new_elem.name); 
				if (cur_lev > -1) stk.push(cur_lev); 
				cur_lev = level+1;  
				normal = 0; 
			}
		}
		
		if ((token =="HLock") || (token =="HBarrier")) {
			string new_line = "int "; int u; int v = 0; 
			while(token_list >> token){
				if (token == "*") new_line += token;
				else {
					u = 0; 
					while (token[u] !='\0'){
						if (token[u] == '*') new_line += "*"; 
						if ((token[u] == '*')||(token[u] == ';')||(token[u] == ',')) token[u]= ' ';
						u++;
					}
					istringstream t_list(token); token = ""; t_list >> token;
					string new_token = "core_"+core+"_"+token;
					if(v > 0) new_line += ", " + new_token;
					else new_line += new_token; 
					v++;
					elem_list[current].variables[token] = new_token;
				}
			}
			new_line += ";"; 
			elem_list[current].vars_lines.push_back(new_line); 
			normal = 0; 
		}
		
		if (token =="HGlobal") {
			token_list >> token;
			string new_line;
			new_line = token + " "; int u; int v = 0; int k = 0; int s = 0; 
			while(token_list >> token){
				if (token == "*") new_line += token;
				else {
					u = 0; 
					while (token[u] !='\0'){
						if (token[u] == '*') new_line += "*"; 
						if ((token[u] == '*')||(token[u] == ';')||(token[u] == ',')) token[u]= ' ';
						u++;
					}
					istringstream t_list(token); token = ""; t_list >> token;
					
					string new_token; 
					if (token.length() > 0){
						if((token[0] == '[') || (token[0] == ']') || (token[0] == '=')){
							k = 1; new_token = token;
							if (token[0] == '=') {
								int l = 0; 
								if((l = line.find("=", l)) != string::npos) {
									new_token = line.substr(l, line.length());
									token_list.str(""); 
									s = 1; 
								}
								
							}
						}
						else new_token = "core_"+core+"_"+token;
						
					}
					else new_token = token;
					
					if((v > 0) && (k == 0)) new_line += ", " + new_token;
					else new_line += new_token; 
					v++;
					elem_list[current].variables[token] = new_token;
				}
			}
			if (s == 0) new_line += ";";  
			elem_list[current].vars_lines.push_back(new_line); 
			normal = 0; 
		}
		
		if (token =="{") {
			level++;
			elem_list[current].lines.push_back(line); 
			normal = 0; 
		}
		
		if (token =="}") {
			elem_list[current].lines.push_back(line);
			normal = 0; 
			if(cur_lev == level) {
				current = find_element(elem_list, elem_list[current].parent); 
				if (current == -1) current = 0;
				if(stk.empty()) cur_lev = -1; 
				else {cur_lev = stk.top(); stk.pop();}
			}
			level--; 
		}
		
		if(normal == 1) elem_list[current].lines.push_back(line);
	}
	
	
	for(int i = 0; i< elem_list.size(); i++){
		for(int j = 0; j< elem_list[i].lines.size(); j++){
			line = elem_list[i].lines[j]; 
			string t_line = line;
			int u = 0; 
			while (line[u] !='\0') {
				if ((line[u] < 48)|| ((line[u] > 57) && (line[u] < 65)) ||
				    ((line[u] > 90) && (line[u] < 95)) || (line[u] > 122))
					line[u]= ' '; 
				u++;
			}
			istringstream token_list(line); 
			map<string, string>::iterator iter;
			while(token_list >> token){
				int k = find_declaration (elem_list, token, elem_list[i].name); 
				if (k != -1){
					iter = elem_list[k].variables.find(token);
					if(iter != elem_list[k].variables.end()) { 
						u = 0; 
						while ((u = t_line.find(iter->first, u)) != string::npos) {
							string str = iter->first; 
							int adv = 0; 
							if (u > 0) {
								if ((t_line[u - 1] == '.') || (letter(t_line[u - 1])) || (letter(t_line[u + str.length()]))){
									u += str.length();
									adv = 1; 
								}
							}
							
							if (adv == 0)  {
								t_line.replace(u, str.length(), iter->second);
								str = iter->second; u += str.length();
							}
						}
					}
				}
			}
			elem_list[i].lines[j] = t_line; 
		}
	}
	
	cout<<"Step 2 done ... \n";
	
	cout<<"Start step 3/3 ... \n";
	
	pos<< "// C/C++ Program \n"; 
	rearrange (elem_list, types); 
	
	for(int j = 0; j< elem_list[0].vars_lines.size(); j++)
		pos<< elem_list[0].vars_lines[j]<<endl;
	pos<<endl;
	
	for(int j = 0; j< elem_list[0].lines.size(); j++)
		pos<< elem_list[0].lines[j]<<endl;
	pos<<endl; 
	
	for(int i = 1; i< elem_list.size(); i++){
		for(int j = 0; j< elem_list[i].vars_lines.size(); j++)
			pos<< elem_list[i].vars_lines[j]<<endl;
		pos<<endl;
	}
	
	for(int i = 1; i< elem_list.size(); i++){
		for(int j = 0; j< elem_list[i].lines.size(); j++)
			pos<< elem_list[i].lines[j]<<endl;
		pos<<endl;
	}
	pos.close();
	
	cout<<"Step 3 done ... \n";
}
