/*  @author : Michel Kinsy
 
 *  Copyright (c) 2010 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
#include <iostream>
#include <cstdlib>
#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <vector>
using namespace std;

// Global variables
string token, line;
char *filename; 
void scanner(void);
void scanner2(void); 
string decimal2hex(int int_num); 
string decimal2Binary(int int_num) ;
string decimal2Binary(int int_num, int width) ;
string binary2hex(string bin_num); 
string int_to_string(int number); 
string hex_string_format(string hex_number) ; 

int core = 0;                // Default memory core  
int stack_size = 1024;       // Default stack frame pointer 
int num_bits_local_mem = 16; // Default local memory size bits

int const MAX_IMMEDIATE = 32767; 

int main (int argc, char * const argv[]) {
    	cout << "Program memory adjustment ................\n";
	
	if (argc < 2) return 1; 
	ostringstream argsout;
	if(argc >= 2) filename = argv[1];
	if(argc == 3) argsout << argv[2];
	if(argc == 4) argsout << argv[2] <<" "<< argv[3];
	if(argc == 5) argsout << argv[2] <<" "<< argv[3] <<" "<< argv[4];
	
	istringstream argsin (argsout.str());
	if(argc > 1) 	argsin >> core;  
	if(argc > 2) 	argsin >> stack_size; 
	if(argc > 3) 	argsin >> num_bits_local_mem; 
	
	scanner();
    	return 0;
}

/* Scanning process and error cheching */
void scanner(void)
{
	string start_str, start;
	string start1 = "0";
	ofstream prog_start_addr; 
	vector <string> instructions;
	int ins_count = 0; 
	ifstream from;
	from.open(filename, ios::in);   /* open filename for reading */
	if(from.is_open())
		cin.rdbuf(from.rdbuf());  /* Redirect cin to the file reader buffer */
	else
	{
		cout << "Unable to open file.\n";  /* Explain the I/O error */
		exit(1);
	}
	
	int i= 0; 
	char new_filename[100]; 
	while ((filename[i] !='.')&&(i<45)) {
		new_filename[i]= filename[i]; i++;
	}
    string core_str = "_c"; core_str.append(int_to_string(core)); 
    for (int j = 0; j < core_str.length(); j++){
        new_filename[i]= core_str[j]; i++; 
    }		
	new_filename[i] = '.'; new_filename[i+1] = 'm'; new_filename[i+2] = 'e'; 
	new_filename[i+3] = 'm'; new_filename[i+4] = '\0';
	
	/* Extracting the lines from the input file */
	int clear = 0; int dispatch = 0; 
	while(getline(cin, line))
	{
		ins_count++; 
		istringstream token_list(line);
		token_list >> token;
		start_str = token; 
		if(clear == 1){
		    if (token[0] =='@') clear = 0; 
		}
		
		token_list >> token; token_list >> token;
		
		if (token == "<Dispatcher>"){
			start = start_str.substr(1,start_str.length()); 
			dispatch = 1; 
		}
		
		if (token == "<Dispatcher1>"){
			start1 = start_str.substr(1,start_str.length());  
		}
		
		if((token == "<main>") && (dispatch == 0)){
			start = start_str.substr(1,start_str.length()); 
		}
		
		if ((token[0] =='<') && (token[1] =='.')) clear = 1; 
		if(clear == 0) instructions.push_back(line); 
	}
	from.close();
	
	// for kernel
	char prog_filename[50]; i=0;
	while ((filename[i] !='.')&&(i<45)) {
		prog_filename[i]= filename[i]; i++;
	}
	
    for (int j = 0; j < core_str.length(); j++){
		prog_filename[i]= core_str[j]; i++; 
    }
	prog_filename[i] = '.'; prog_filename[i+1] = 't'; prog_filename[i+2] = 'x';
	prog_filename[i+3] = 't'; prog_filename[i+4] = '\0';
	
 	string stack_str = hex_string_format(decimal2hex(stack_size));
    int local_mem = core << num_bits_local_mem; 
    int start_address = local_mem + (ins_count*4); 
	instructions.push_back("@"+ decimal2hex(ins_count));
	string thread_start0 = decimal2hex(ins_count);
	
	for (int i = 0; i < 32; i++) {
	    if (i < 16) instructions.push_back("240"+decimal2hex(i)+"0000");
	    if (i >= 16 && i < 29) 
	    	instructions.push_back("24"+decimal2hex(i)+"0000");
		
		if (i == 29) {
			instructions.push_back("241d"+hex_string_format(decimal2hex(core)));
			string sll_str = binary2hex("1"+decimal2Binary(num_bits_local_mem, 5)+"00"); 
			instructions.push_back("001de"+sll_str+"0");
			instructions.push_back("27bd"+stack_str); 
		}
		
		if (i == 30) instructions.push_back("241e0000");
		if (i == 31) instructions.push_back("241f0000");
	}
	ins_count+=32;
	
	string tmp =""; 
	tmp +="0c";
	int len = 6 - start.length(); 
	for (int i = 0; i < len; i++) tmp +="0";
	tmp +=start;
	instructions.push_back(tmp);
	
	// for buffering 
	for (int i = 0; i < 16; i++) 
		instructions.push_back("00000000");
	instructions.push_back("0c000000");
	ins_count+=20;
	
	//Second hardware thread set up
	string stack_str1 = hex_string_format(decimal2hex(stack_size/2));
	instructions.push_back(""); 
	instructions.push_back("@"+decimal2hex(ins_count));
	string thread_start1 = decimal2hex(ins_count); 
	for (int i = 0; i < 32; i++) {
	    if (i < 16) instructions.push_back("240"+decimal2hex(i)+"0000");
	    if (i >= 16 && i < 29) 
	    	instructions.push_back("24"+decimal2hex(i)+"0000");
		
		if (i == 29) {
			instructions.push_back("241d"+hex_string_format(decimal2hex(core)));
			string sll_str = binary2hex("1"+decimal2Binary(num_bits_local_mem, 5)+"00"); 
			instructions.push_back("001de"+sll_str+"0");
			instructions.push_back("27bd"+stack_str1); 
		}
		
		if (i == 30) instructions.push_back("241e0000");
		if (i == 31) instructions.push_back("241f0000");
	}
	
	tmp ="0c";
	len = 6 - start1.length(); 
	for (int i = 0; i < len; i++) tmp +="0";
	tmp +=start1;
	instructions.push_back(tmp);
	
	// for buffering 
	for (int i = 0; i < 16; i++)
		instructions.push_back("00000000");
	instructions.push_back("0c000000");
	
	bool default_start = true; 
	if ((int)instructions.size() > 8) {
		for(int i =1; i < 8; i++) {
			istringstream token_list(instructions[i]);
			token_list >> token;
			if (token !="@0" && token !="00000000"){
				default_start = false; 
				break; 
			}
			else {
				token_list >> token; token_list >> token;
				if (token == "00000010") {
					string tmp2 = thread_start0; 
					len = 6 - tmp2.length(); 
					tmp ="0c";
					for (int j = 0; j < len; j++) tmp +="0";
					tmp +=tmp2;
					instructions[i] = tmp+"  // 00000010 j	"+tmp2; 
				}
				if (token == "00000014") {
					string tmp2 = thread_start1; 
					len = 6 - tmp2.length(); 
					tmp ="0c";
					for (int j = 0; j < len; j++) tmp +="0";
					tmp +=tmp2;
					instructions[i] = tmp+"  // 00000014 j	"+tmp2; 
				}	
			} 
		}
	}
	else default_start = false; 

	ofstream pos (new_filename, ios::out); 
	for(int i =0; i < (int)instructions.size(); i++) 
		pos<<instructions[i]<<endl; 
	pos<<endl; 
	pos.close();
	
	prog_start_addr.open (prog_filename); 
	prog_start_addr<<"============================================================\n";
	prog_start_addr<<"Core: "<<core<<"\n";
	if (default_start) 
		prog_start_addr<<"Program Starting Address: 'h"<<decimal2hex(local_mem + 16)<<"\n";
	else 
		prog_start_addr<<"Program Starting Address: 'h"<<decimal2hex(start_address)<<"\n";
	
	int m_addr = local_mem + (strtol(start.c_str(), NULL, 16) * 4); 
	prog_start_addr<<"Main Function Address: 'h"<<decimal2hex(m_addr)<<"\n";
	prog_start_addr<<"Stack Frame Size: 'h"<<stack_str<<"\n";
	prog_start_addr<<"============================================================\n";
	prog_start_addr.close(); 
}


string decimal2hex(int int_num) 
{
	int hex_values[8]; 
	int remainder = int_num;
	int index = 0;
	string hex_str = "";  
	
	while (remainder>15) {
		hex_values[index] = remainder% 16; 
		remainder = remainder/ 16; 
		index ++; 
	}
	hex_values[index] = remainder; 
	
	while (index >= 0) {
		if (hex_values[index]==0) hex_str.append("0");
		else if (hex_values[index]==1) hex_str.append("1");
		else if (hex_values[index]==2) hex_str.append("2");
		else if (hex_values[index]==3) hex_str.append("3");
		else if (hex_values[index]==4) hex_str.append("4");
		else if (hex_values[index]==5) hex_str.append("5");
		else if (hex_values[index]==6) hex_str.append("6");
		else if (hex_values[index]==7) hex_str.append("7");
		else if (hex_values[index]==8) hex_str.append("8");
		else if (hex_values[index]==9) hex_str.append("9");
		else if (hex_values[index]==10) hex_str.append("a");
		else if (hex_values[index]==11) hex_str.append("b");
		else if (hex_values[index]==12) hex_str.append("c");
		else if (hex_values[index]==13) hex_str.append("d");
		else if (hex_values[index]==14) hex_str.append("e");
		else if (hex_values[index]==15) hex_str.append("f");
		index --; 
	}
	return hex_str;
}

string decimal2Binary(int int_num) 
{
	int bin_values[32]; 
	int remainder = int_num;
	int index = 0;
	string bin_str = "";  
	
	while (remainder>1) {
		bin_values[index] = remainder% 2; 
		remainder = remainder/ 2; 
		index ++; 
	}
	bin_values[index] = remainder; 
	
	while (index >= 0) {
		if (bin_values[index]==0) bin_str.append("0");
		else if (bin_values[index]==1) bin_str.append("1");
		index --; 
	}
	return bin_str;
}

string decimal2Binary(int int_num, int width) 
{
	int bin_values[32]; 
	int remainder = int_num;
	int index = 0;
	string bin_str = "";  
	
	while (remainder>1) {
		bin_values[index] = remainder% 2; 
		remainder = remainder/ 2; 
		index ++; 
	}
	bin_values[index] = remainder; 
	for (int i = 1; i< (width - index); i++) bin_str.append("0");
	
	while (index >= 0) {
		if (bin_values[index]==0) bin_str.append("0");
		else if (bin_values[index]==1) bin_str.append("1");
		index --; 
	}
	return bin_str;
}

string binary2hex(string bin_num) 
{
	string hex_str = "";  
	int length = bin_num.length(); 
	int adj    = length% 4;
	string cur; 
	
	string temp = ""; 
	if (adj == 3){ temp.append("0"); temp.append(bin_num); }
	if (adj == 2){ temp.append("00"); temp.append(bin_num); }
	if (adj == 1){ temp.append("000"); temp.append(bin_num); }
	if (adj == 0){temp.append(bin_num); }
		
	int index = 0; 
	while (index < temp.length()) {
		cur = temp.substr (index,4);
		if (cur == "0000") hex_str.append("0");
		else if (cur == "0001") hex_str.append("1");
		else if (cur == "0010") hex_str.append("2");
		else if (cur == "0011") hex_str.append("3");
		else if (cur == "0100") hex_str.append("4");
		else if (cur == "0101") hex_str.append("5");
		else if (cur == "0110") hex_str.append("6");
		else if (cur == "0111") hex_str.append("7");
		else if (cur == "1000") hex_str.append("8");
		else if (cur == "1001") hex_str.append("9");
		else if (cur == "1010") hex_str.append("a");
		else if (cur == "1011") hex_str.append("b");
		else if (cur == "1100") hex_str.append("c");
		else if (cur == "1101") hex_str.append("d");
		else if (cur == "1110") hex_str.append("e");
		else if (cur == "1111") hex_str.append("f");
		index += 4; 
	}
	
	return hex_str;
}

string hex_string_format(string hex_number) 
{
	string hex_format_str = "";  
	int length = hex_number.length(); 
	if (length > 4) {
        hex_format_str = "7fff"; // default max 'h7fff
    }
	else {
		string temp = ""; 
		if (length == 4){temp.append(hex_number); }
		if (length == 3){ temp.append("0"); temp.append(hex_number); }
		if (length == 2){ temp.append("00"); temp.append(hex_number); }
		if (length == 1){ temp.append("000"); temp.append(hex_number); }
		hex_format_str = temp; 
    } 
	return hex_format_str;
}

string int_to_string(int number)
{
	stringstream str;
	str << number;
	return str.str();
}
